# -*- coding: utf-8 -*-
from sspymgr import db, createLogger
from sqlalchemy.sql import text
from base64 import b64decode

logger = createLogger('plugin_settings', stream=False, logger_prefix="[Plugin Settings]")

class WebguiSetting( db.Model ):
    """
    Type (can be ignored, just used as utility):
        String: Value type by default, should be processed by other logic
        Image: Value are encoded with base64 format
        Number: Treat value as number( use float builtin function), 
        Boolean: are stored in number with only two candicates (0 or 1)
        Json: Value are encoded with json format
        Hidden: Not visible for even manager
    """
    __tablename__ = 'webguiSetting'
    id = db.Column( db.Integer, primary_key = True, autoincrement = True )
    key = db.Column( db.String( 255 ), unique = True )
    value = db.Column( db.Text )
    type = db.Column( db.String( 32 ), default = 'String' )

    def getTypedValue( self ):
        if not self.type:
            return self.value
        if self.type == 'Boolean':
            return self.value == '1'
        if self.type == 'Number':
            try:
                return int( self.value )
            except Exception:
                return float( self.value )
        return self.value
    
    @staticmethod
    def count():
        sql = 'SELECT count(*) FROM %s' % WebguiSetting.__tablename__
        res = db.engine.execute( text( sql ) ).first()
        return res[ 0 ]

    @staticmethod
    def getSetting(key, default_value=None, default_type="String"):
        setting = WebguiSetting.query.filter_by(key=key).first()
        if setting is None and default_value is not None :
            setting = WebguiSetting(key=key, value=default_value, type=default_type)
            db.session.add(setting)
            db.session.commit()
        return setting

    def to_dict(self):
        di = {
            'id': self.id,
            'key': self.key,
            'value': self.value if type(self.value) is str else self.value.decode() ,
            'type': self.type,
        }
        return di

def init_webgui_setting():
    """
    alipay_enabled: bool default false
    wechatpay_enabled: bool default false
    alipay_qrcode: image data with base64 encode
    wechatpay_qrcode: image data with base64 encode
    signup_enabled: bool default true
    signup_without_invitation: int  default 8
    signup_email_limit: int  default 2
    shadowsock_port_range_start: Number default '50000'
    """
    if WebguiSetting.count() != 0:
        return
    settings = [
        ( "alipay_enabled",    "0", "Boolean" ),
        ( "wechatpay_enabled", "0", "Boolean" ),
        ( "signup_enabled", "1", "Boolean" ),
        ( "signup_without_invitation", "8", "Number" ),
        ( "signup_email_limit", "2", "Number" ),
        ( "shadowsock_port_range_start", "45000", "Number" ),
    ]
    for setting in settings:
        wgs = WebguiSetting( key = setting[0], value = setting[1], type = setting[2] )
        db.session.add( wgs )
    
    db.session.commit()

def init(app):
    app.m_events.on('beforeRegisterApi', registerApi)
    init_webgui_setting()
    logger.debug("inited")

from flask import jsonify

def registerApi(api):
    # TODO change setting item
    @api.route_admin('/setting/getAll', methods=['POST'])
    def get_all_settings():
        all = WebguiSetting.query.all()
        settings = [s.to_dict() for s in all]
        # for s in all:
        #     settings.append(s.to_dict())
        return jsonify({
            'status': 'success',
            'setting': settings
        })
    logger.debug("api registered.")


# -*- coding: utf-8 -*-

import re
EMAIL_REGEX = re.compile(r'[^@]+@[^@]+\.[^@]+')

from .path_helper import createLogger

logger = createLogger("sspymgr")

from ._plugin import Events, BackTask
events = Events()
tasker = BackTask(events)

from .configuration import Configuration
config = Configuration()

from .mail import EmailManager
emailManager = EmailManager(config)

from .sscontroller import SSController
controller = SSController()



# -*- coding: utf-8 -*-

from sspymgr.models import db
from datetime import datetime, timedelta


class Account(db.Model):
    """
    Represents shadowsocks server account, port and password can not be None
    """
    __tablename__ = 'account'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    # corresponding with Table User, but can be created by manager
    userId = db.Column(db.Integer, unique=True)  
    data = db.Column(db.String(255))
    port = db.Column(db.Integer, nullable=False)
    password = db.Column(db.String(255), nullable=False)
    totalFlow = db.Column(db.BigInteger)
    expire = db.Column(db.DateTime)
    server = db.Column(db.String(255))
    method = db.Column(db.String(64), default='aes-256-cfb')
    # server = db.Column( db.String( 255 ) )  # no need for multi server now
    """
    @field type
        0 registered by email and have a 
    """
    type = db.Column(db.Integer)
    status = db.Column(db.String(32))
    createTime = db.Column(db.DateTime, default=datetime.now())

    def __init__(self, **kwargs):
        super(Account, self).__init__(**kwargs)

    def get_status(self):
        if self.status == 0:
            return 0
        return 0

    def is_valid(self):
        now = datetime.now()
        if now > self.expire:
            return False
        # accountFlow = AccountFlow.query.filter_by( accountId = self.id ).first()
        flow = self._flow.flow
        if flow > self.totalFlow:
            return False
        return True

    def clear_flow(self):
        self._accountFlow.flow = 0
        self._flow.flow = 0
        db.session.commit()

    def get_used_flow(self):
        if hasattr(self, '_flow'):
            return self._flow.flow
        return None

    def __repr__(self):
        return '<Account %d>' % self.id

    def to_dict(self):
        di = {
            'id': self.id,
            'userId': self.userId,
            'data': self.data,
            'port': self.port,
            'password': self.password,
            'totalFlow': self.totalFlow,
            'usedFlow': self.get_used_flow(),
            'server': self.server,
            'method': self.method,
            'type': self.type,
            'status': self.status,
            'expire': self.expire.timestamp(),
            'createTime': self.createTime.timestamp(),
        }
        return di


class AccountFlow(db.Model):
    __tablename__ = 'accountFlow'

    port = db.Column(db.Integer, primary_key=True)
    flow = db.Column(db.BigInteger, default=0)  # flow used
    status = db.Column(db.String(255), default='checked')
    accountId = db.Column(db.Integer, unique=True)
    updateTime = db.Column(
        db.DateTime, default=datetime.now)  # means the account last connected
    autoBanTime = db.Column(db.DateTime)

    # serverId  = db.Column( db.Integer )
    checkTime = db.Column(db.DateTime)
    nextCheckTime = db.Column(db.DateTime)

    def consume_flow(self, flow: int):
        self.flow += flow
        self.updateTime = datetime.now()

    def __repr__(self):
        return '<Flow port: %d, flow: %d>' % (self.port, self.flow)

    def to_dict(self):
        di = {
            'port': self.port,
            'flow': self.flow,
            'status': self.status,
            'accountId': self.accountId,
            'updateTime': self.updateTime.timestamp(),
            'autoBanTime': self.autoBanTime.timestamp(),
            'checkTime': self.checkTime.timestamp(),
            # 'nextCheckTime': self.nextCheckTime.timestamp(),
        }
        return di


class Command(db.Model):
    __tablename__ = 'command'
    id = db.Column(db.Integer, primary_key=True)
    code = db.Column(db.String(255))
    time = db.Column(db.String(255))




# -*- coding: utf-8 -*-
from sspymgr.models import db
from datetime import datetime, timedelta

# class SaveFlow( db.db.Model ):
#     __tablename__ = 'saveFlow'
#     id = db.db.Column( db.db.Integer, primary_key = True, autoincrement = True )
#     port = db.db.Column( db.db.Integer )
#     flow = db.db.Column( db.db.BigInteger )
#     time = db.db.Column( db.db.DateTime )
#     accountId = db.db.Column( db.db.Integer )

# def genFlowRecord( tablename ):
#     def __repr__( self ):
#         return '<FlowRecord: %s>' % tablename

#     props = {
#         '__tablename__' : tablename,
#         'id' : db.Column( db.Integer, primary_key = True, autoincrement = True ),
#         'flow' : db.Column( db.BigInteger ),
#         'port' : db.Column( db.Integer ),
#         'time' : db.Column( db.DateTime ),
#         'accountId' : db.Column( db.Integer ),
#     }
#     newclass = type( tablename, ( db.Model, ), props )
#     return newclass

# FlowRecord = genFlowRecord( 'flowRecord' )
# FlowRecord5Min = genFlowRecord( 'flowRecord5min' )
# FlowRecordHour = genFlowRecord( 'flowRecordHour' )
# FlowRecordDay  = genFlowRecord( 'flowRecordDay' )


class FlowRecord5Min(db.Model):
    __tablename__ = 'flowRecord_5min'
    id = db.Column(db.Integer, primary_key=True)
    port = db.Column(db.Integer)
    flow = db.Column(db.Integer)  # unit of bytes
    date = db.Column(
        db.String, nullable=False)  # store date in the format: 2018-10-10
    hour = db.Column(db.Integer, nullable=False)  # 0 - 23
    minute = db.Column(
        db.Integer, nullable=False
    )  # noraml minute here, gap is 5, from 0 to 55: 0 ==> [ 00:00 ~ 04:59 ]

    @staticmethod
    def genRecord(port: int, flow: int, time: datetime):
        """time is record time, but should be 
        """
        d = '{:%Y-%m-%d}'.format(time)
        record = FlowRecord5Min(
            port=port, flow=flow, date=d, hour=time.hour, minute=time.minute)
        db.session.add(record)

    def to_dict(self):
        di = {
            'flow': self.flow,
            'hour': self.hour,
            'minute': self.minute
        }
        return di

    def __repr__(self):
        return '<FlowRecord5Min %d: %d at %s %d>' % (self.port, self.flow,
                                                     self.date, self.hour)


from sqlalchemy import func
from sqlalchemy.sql import label


class FlowRecordDay(db.Model):
    __tablename__ = 'flowRecord_Day'
    id = db.Column(db.Integer, primary_key=True)
    port = db.Column(db.Integer)
    flow = db.Column(db.Integer)  # unit of bytes
    date = db.Column(
        db.String,
        nullable=False)  # store date in the format (year and month): 2018-10
    day = db.Column(db.Integer, nullable=False)  # store day, from 1 to 31

    def to_dict(self):
        di = {
            'flow': self.flow,
            'day': self.day,
        }
        return di

    def __repr__(self):
        return '<FlowRecordDay %d: %d at %s %d>' % (self.port, self.flow,
            self.date, self.day)

    @staticmethod
    def genRecord(port: int, flow: int, time: datetime):
        """time is record time, but should be 
        """
        date = '{:%Y-%m}'.format(time)
        record = FlowRecordDay(port=port, flow=flow, date=date, day=time.day)
        db.session.add(record)

from .utils import logger
class FlowStats(object):
    """统计流量，将流量记录到数据库中
    """

    def __init__(self, *args, **kwargs):
        self.min = {}
        # self.hour = {}
        self.day = {}
        return super().__init__(*args, **kwargs)

    def recordFlow(self, accFlow: dict):
        """将流量数据记录到 dict 中
        accFlow: { 
            45001: {'flow': 12, 'status': 'added' }, 
            45002: {...}  
        }
        """
        for port, detail in accFlow.items():
            if port not in self.min:
                self.min[port] = detail['flow']
            else:
                self.min[port] = self.min[port] + detail['flow']
        # logger.debug("5 min: {}".format(self.min))
        for port in self.min:
            if port not in self.day:
                self.day[port] = self.min[port]
            else:
                self.day[port] += self.min[port]

    def record5minToDb(self):
        # logger.debug("5 min record: {}".format(str(self.min)))
        for port in self.min:
            if self.min[port] == 0:
                # no need to record
                continue
            FlowRecord5Min.genRecord(port, self.min[port], datetime.now())
            self.min[port] = 0
        db.session.commit()

    def record1hourToDb(self):
        pass

    def record1dayToDb(self):
        # FlowRecordDay.genRecordFrom5min( now )
        for port in self.day:
            if self.day[port] == 0:
                continue
            FlowRecordDay.genRecord(port, self.day[port], datetime.now())
            self.day[port] = 0
        db.session.commit()

from datetime import datetime, timedelta

from sspymgr import formatTime
from sspymgr.models import db
from sspymgr.sscontroller import Account, AccountFlow
from sspymgr.globalvars import emailManager
from web_user import User, UserType
import logging
logger = logging.getLogger("plugin_sscontroller")

def checkAllAccounts():
    """Action: check accounts status and userType
    """
    dbaccounts = Account.query.all()
    # accountFlow = AccountFlow.query.all()
    for acc in dbaccounts:
        # for flow in accountFlow:
        #     if acc.port == flow.port:
        #         acc._flow = flow
        #         break
        # if acc.is_valid():
        uid = acc.userId
        user = User.query.filter_by(id=uid).first()
        if user is None:
            continue
        now = datetime.now()
        if now < acc.expire: 
            if canSendLastDay(user) and (acc.expire - now) < timedelta(days=1):
                sendLastDay(acc)
        else:
            user.type = UserType.INVALID.value
            db.session.commit()
    

def sendLastDay(acc: Account, user: User):
    logger.debug("<{} : {}> will expire soon".format(user.email, user.comment))

    content = """Hi, your account of SSPY-MGR will be expired within one day (at {}).<br />
    Please go to {} to rechart Your account.(Your ss account Port: {})
    """.format(formatTime(acc.expire), acc.server, acc.port)
    user.type = UserType.NOTIFIED.value
    subject = "账户过期提醒"

    emailManager.add_email(
        to=user.email,
        subject=subject,
        content=content,
        type="last day notify")


def canSendLastDay(user):
    if user.type is None or len(user.type) == 0:
        return False
    utype = UserType(user.type)
    if (utype == UserType.BANNED or utype == UserType.TRIAL):
        return False
    if (utype != UserType.NOTIFIED):
        return True
    return False


# -*- coding: utf-8 -*-
from sspymgr import getRandomCode, convertFlowToByte
from sspymgr.models import db
from sspymgr.sscontroller import *
from sspymgr.globalvars import controller
import logging
logger = logging.getLogger("plugin_sscontroller")


from datetime import datetime, timedelta
from flask import request, jsonify, session

from web_user import User, UserType
from sqlalchemy.sql import func

from .flow_stats import FlowRecord5Min, FlowRecordDay

from web_settings import WebguiSetting
def availiable_port() -> int:
    port = db.session.query(func.max(Account.port)).scalar()
    # db.session.query(db.func.max(User.numLogins)).scalar()
    # maxsql = 'SELECT max(port) FROM %s' % Account.__tablename__
    # max = db.engine.execute( text( maxsql ) ).first()
    # if max[0] is None:
    if port is None:
        port = WebguiSetting.getSetting(key='shadowsock_port_range_start',\
            default_value=45000, default_type="Number").getTypedValue()
        db.session.commit()
    else:
        port += 1
    return port


def registerApi(api):
    """注册路由
    """
    @api.route_admin('/user/getDetail', methods=['GET', 'POST'])
    def admin_getUserDetail():
        userId = request.form.get('userId')
        user = User.query.filter_by(id=userId).first()
        state = {'status': 'success'}
        if not user:
            state['status'] = 'fail'
            return jsonify(state)
        state['user'] = user.to_dict()
        account = Account.query.filter_by(userId=userId).first()
        if not account:
            state['account'] = {}
            return jsonify(state)
        state['account'] = account.to_dict()
        accFlow = AccountFlow.query.filter_by(port=account.port).first()
        state['accountFlow'] = accFlow.to_dict()
        return jsonify(state)


    @api.route_admin('/account/getDetail', methods=['POST'])
    def get_admin_accountDetail():
        id = request.form.get('id', None)
        state = {'status': 'success'}
        if id is None:
            state['status'] = 'fail'
            return jsonify(state)
        acc = Account.query.filter_by(id=id).first()
        state['account'] = acc.to_dict()
        accFlow = AccountFlow.query.filter_by(port=acc.port).first()
        state['accountFlow'] = accFlow.to_dict()
        if acc.userId:
            user = User.query.filter_by(id=acc.userId).first()
            state['user'] = user.to_dict()
        else:
            state['user'] = {}
        
        address = "{}:{}@{}:{}".format(acc.method, acc.password, acc.server, acc.port)
        state['ssAddress'] = address
        return jsonify(state)

    @api.route_admin('/account_port', methods=['POST'])
    def on_get_account_port():
        port = availiable_port()
        return str(port)

    @api.route_admin('/account/get', methods=['POST'])
    def get_admin_account():
        accounts = controller.list_account()
        # accounts = Account.query.filter().all()
        return jsonify({
            'status': 'success',
            'accounts': [a.to_dict() for a in accounts]
        })

    @api.route_admin('/accountAdd', methods=['POST'])
    def add_account():
        port = request.form.get('port')
        try:
            port = int(port)
        except Exception:
            port = availiable_port()  # should get a useable port

        password = request.form.get('password')

        flow = request.form.get('flow')  # in the unit GB
        try:
            flow = int(flow)
        except:
            flow = 10
        flow = convertFlowToByte(flow, 'G')  # in the unit byte

        duration = request.form.get('duration')
        try:
            duration = int(duration)
        except Exception:
            duration = 7
        if not port or not password or not flow or not duration:
            return 'Bad request'
        expire = timedelta(days=duration) + datetime.now()
        controller.add_account(port=port, password=password, flow=flow, expire=expire)
        logger.info('successfully add account %d' % port)
        return 'success'
    
    @api.route_admin('/account/updatePassword', methods=['POST'])
    def on_account_update_password():
        newpasswd = request.form.get('password', None)
        port = request.form.get('port', None)
        if newpasswd is None or port is None:
            return jsonify('fail')
        port = int(port)
        if controller.update_password(newpasswd, port=port):
            return jsonify('success')
        else:
            return jsonify('fail')


    @api.route_admin('/account/updateTotalFlow', methods=['POST'])
    def on_account_update_totalflow():
        port = request.form.get('port', None)
        flow = request.form.get('flow', None)  # here the flow is in unit of Gb
        unit = request.form.get('flowUnit', 1024 * 1024 * 1024)
        try:
            port = int(port)
            flow = int(flow)
            unit = int(unit)
        except Exception:
            return jsonify('fail')
        account = Account.query.filter_by(port=port).first()
        if account is None:
            return jsonify('fail')

        flow = flow * unit
        account.totalFlow = flow
        
        db.session.commit()
        return jsonify('success')


    @api.route_admin('/account/randomPassword', methods=['POST'])
    def on_account_random_password():
        port = request.form.get('port', None)
        result = {'status': 'fail'}
        if port is None:
            return jsonify(result)
        port = int(port)
        newpasswd = getRandomCode(8)
        if controller.update_password(newpasswd, port=port):
            result['status'] = 'success'
            result['password'] = newpasswd
        else:
            result['status'] = 'fail'
        return jsonify(result)


    @api.route_admin('/account/updateExpiration', methods=['POST'])
    def on_account_update_expiration():
        result = {'status': 'fail'}
        port = request.form.get('port', None)
        days = request.form.get('days', None)
        try:
            port = int(port)
            days = int(days)
        except Exception:
            return jsonify(result)
        result['status'] = 'success'
        acc = Account.query.filter_by(port=port).first()
        if not acc:
            result['status'] = 'fail'
            return jsonify(result)
        td = timedelta(days=days)
        acc.expire += td
        if acc.expire > datetime.now():
            user = User.query.filter_by(id = acc.userId).first()
            if user is not None:
                user.type = UserType.ACTIVE.value
            if acc.totalFlow == 0:
                setting = WebguiSetting.getSetting("default_account_flow",\
                    default_value=1, default_type="Number")
                acc.totalFlow = setting.getTypedValue() * 1024 * 1024 * 1024
            
        db.session.commit()
        result['nexpire'] = acc.expire.timestamp()
        return jsonify(result)

    @api.route_admin('/flowUsage/getInDay', methods=['GET', 'POST'])
    def admin_getFlowUsage_day():
        state = {'status': 'success'}
        accountId = request.form.get('accountId', -1)
        account = Account.query.filter_by(id = accountId).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            d = int(date[2])
            now = datetime(year=y, month=m, day=d)
        except:
            now = datetime.now()
        date = '{:%Y-%m-%d}'.format(now)
        records = FlowRecord5Min.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)
        
    @api.route_admin('/flowUsage/getInMon', methods=['GET', 'POST'])
    def admin_getFlowUsage_month():
        state = {'status': 'success'}
        accountId = request.form.get('accountId', -1)
        account = Account.query.filter_by(id = accountId).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            now = datetime(year=y, month=m)
        except:
            now = datetime.now()
        date = '{:%Y-%m}'.format(now)
        records = FlowRecordDay.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)

    @api.route_admin('/account/<id>/flowRecord_5min', methods=['GET', 'POST'])
    def get_account_flowRecord_5min(id):
        """ flow unit is KB
        """
        acc = Account.query.filter_by(id=id).first()
        if not acc:
            return '{}'

        now = datetime.now()
        date = '{:%Y-%m-%d}'.format(now)
        records = FlowRecord5Min.query.filter_by(port=acc.port, date=date).all()
        result = {}
        for r in records:
            key = r.hour * 60 + r.minute
            result[key] = r.flow / 1000
        return jsonify(result)

    @api.route_user('/flowUsage/getInDay', methods=['GET', 'POST'])
    def user_getFlowUsage_day():
        state = {'status': 'success'}
        uid = session.get('userid')
        account = Account.query.filter_by(userId=uid).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            d = int(date[2])
            now = datetime(year=y, month=m, day=d)
        except:
            now = datetime.now()
        date = '{:%Y-%m-%d}'.format(now)
        records = FlowRecord5Min.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)
        
    @api.route_user('/flowUsage/getInMon', methods=['GET', 'POST'])
    def user_getFlowUsage_month():
        state = {'status': 'success'}
        uid = session.get('userid')
        account = Account.query.filter_by(userId=uid).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            now = datetime(year=y, month=m)
        except:
            now = datetime.now()
        date = '{:%Y-%m}'.format(now)
        records = FlowRecordDay.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)

    @api.route_user('/accountDetail', methods=['POST'])
    def user_accountDetail():
        uid = session['userid']
        account = Account.query.filter_by(userId=uid).first()
        result = {'status': 'fail'}
        if account is not None:
            accountFlow = AccountFlow.query.filter_by(port=account.port).first()
            account._flow = accountFlow
            server = request.host
            if ":" in server:
                server = server.split(':')[0]
            enabled = (account.status == 'added')
            account.server = account.server or server
            account.last_connected = accountFlow.updateTime

            # 用于生成二维码的地址
            # address = '-'.join(
            #     [account.server,
            #     str(account.port), account.password])
            address = "{}:{}@{}:{}".format(account.method, account.password, account.server, account.port)

            result['account'] = account.to_dict()
            result['ssAddress'] = address
            result['enabled'] = enabled
            result['status'] = 'success'
        return jsonify(result)


    @api.route_user('/account/updatePassword', methods=['POST'])
    def on_user_update_account_password():
        password = request.form.get('password')
        if not password:
            return jsonify('failed')
        uid = session['userid']
        return 'success' if controller.update_password(password, uid=uid) else 'fail'

    logger.info("routes api registered")




# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from sspymgr.path_helper import createLogger
logger = createLogger('sscontroller', stream=False, logger_prefix="[Core SSController]")

import sys, os
def ssAddr(addr2str = False):
    """获取统一的 SS 通信交互地址
    """
    if sys.platform == 'win32':
        mgr_addr = ('127.0.0.1', 6001)
    elif sys.platform.startswith( 'linux' ):
        mgr_addr = '/var/run/shadowsocks-manager.sock'
        try:
            os.unlink( mgr_addr )
        except Exception:
            pass

    # some problem in unix sock file, using tcp mode always
    mgr_addr = ('127.0.0.1', 6001)
    if addr2str:
        mgr_addr = '{}:{}'.format( mgr_addr[0], mgr_addr[1])
    return mgr_addr
    


# -*- coding: utf-8 -*-
from .path_helper import DATA_DIR
from os.path import abspath, join

_SQLITE_DATABASE = 'database.db'
SQLITE_DATABASE_LOC = abspath( join( DATA_DIR, _SQLITE_DATABASE ) )

"""
from ..globalvars import app

app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///%s' % SQLITE_DATABASE_LOC
# app.config['SQLALCHEMY_COMMIT_ON_TEARDOWN'] = True
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True

from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy( app )
"""


### for pagenate

from sqlalchemy.orm import  Query
import math

def paginate(self, page=None, per_page=None, to_dict=True):
    """
    分页函数
    :param self:
    :param page:
    :param per_page:
    :return:
    """
    if page is None:
        page = 1

    if per_page is None:
        per_page = 20

    items = self.limit(per_page).offset((page - 1) * per_page).all()

    if not items and page != 1:
        return {'total': 0, 'page': page, 'error': 'no such items'}
        
    if page == 1 and len(items) < per_page:
        total = len(items)
    else:
        total = self.order_by(None).count()
    
    if to_dict:
        ditems = [item.to_dict() for item in items]
    else:
        ditems = items

    return {
        'page': page, 
        'per_page': per_page, 
        'total': total, 
        'items': ditems
    }
    # return Pagination(self, page, per_page, total, items)


from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy as sqlalchemy

def createDatabase():
    Model = declarative_base()

    engine = sqlalchemy.create_engine('sqlite:///{}'.format(SQLITE_DATABASE_LOC))
    Model.metadata.bind = engine
    session = orm.scoped_session(orm.sessionmaker(bind = engine))

    sqlalchemy.session = session
    sqlalchemy.Model = Model
    sqlalchemy.engine = engine

    sqlalchemy.create_all = Model.metadata.create_all
    Query.paginate = paginate
    Model.query = session.query_property()
    return sqlalchemy

db = createDatabase()



# -*- coding: utf-8 -*-
"""负责底层操作，如：与 shadowsock 服务器交互
"""

import socket 
# from gevent import socket, spawn, joinall
# from gevent.pool import Pool
import os, sys
DEFAULT_METHOD = "aes-256-cfb"
from .utils import logger
# from .database import db, Account, Command

class SSProtocol( object ):
    """only generate corresponding string, not communicate with shadowsocks server
    """
    def ping( self ):
        return b'ping', b'pong'

    def add_port( self, port, password, method ):
        cmd = 'add: {{ "server_port": {:d}, "password": "{:s}", "method": "{:s}" }}'.format( port, password, method)
        # command = Command(code=cmd)
        # db.session.add
        return cmd.encode( 'ascii' ), b'ok', True, port
    
    def remove_port( self, port ):
        cmd = "remove: {{ 'server_port': {:d} }}".format(port)
        return cmd.encode( 'ascii' ), b'ok', False, port

# import re
# STAT = re.compile( b'stat: {[^s]+}' )
import json
from queue import Queue


class SSAdapter( object ):
    """contains two queues, one of which contains commands to be sent,
    one of which contains string recieved
    """

    def __init__( self, stater ):
        self.stater = stater
        self.queue = Queue()
        self.protocol = SSProtocol()
        self.accounts = dict() # key is port: int, value is another dict
        self.stat_count = 0

    def conn_ser( self, server ):
        if type( server ) is not tuple:
            from threading import current_thread
            name = current_thread().name
            sock_file = '/tmp/sspymgr_control_%s.sock'  % name  # if more than one thread using same file, problems will definately occure
            try:
                os.unlink( sock_file )
            except Exception as exc:
                # print( 'error: ', exc )
                pass

            self._socket = socket.socket( socket.AF_UNIX, socket.SOCK_DGRAM )
            self._socket.bind( sock_file )
        else:
            self._socket = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )  ## UDP connection

        self._socket.connect( server )
        # self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.settimeout( 0.1 )  # never timeout
        self._socket.setblocking( True ) # block
        self.ping()


    def ping( self ):
        cmd = self.protocol.ping()
        self.queue.put( cmd )

    def add_port( self, port: int, password: str, method=DEFAULT_METHOD ):
        """add processing status: created -> added
        """
        self.accounts[ port ] = { 'flow': 0, 'status': 'created' }

        cmd = self.protocol.add_port(port, password, method)
        # logger.debug("add port: {}".format(cmd))
        self.queue.put( cmd )
    
    def remove_port( self, port: int ):
        if port not in self.accounts:
            return
        self.accounts[ port ] = { 'flow': 0, 'status': 'removing' }
        cmd = self.protocol.remove_port(port)
        # self.accounts.pop( port ) 
        self.queue.put( cmd )
    
    def update_port( self, port: int, password: str, method = DEFAULT_METHOD):
        self.remove_port( port )
        self.add_port( port, password, method )

    def contains_port( self, port: int ):
        return port in self.accounts

    def execute(self):
        cmd = None
        if not self.queue.empty():
            cmd = self.queue.get()
            # logger.debug("sending: {}".format(cmd))
            self._socket.send( cmd[0] )
        try:
            rmsg = self._socket.recv( 2048 )
            self._processResponse(rmsg, cmd)
        except Exception as exc:
            logger.warn("cmd: {}".format(exc) )
            pass

        self.stat_count += 1

        if( self.stat_count == 60 ):
            """avoid long time not recved stats information
            """
            self.ping()
            self.stater( self.accounts, force=True )
            self.stat_count = 0

    def _processResponse(self, rmsg, cmd=None):
        """stats is a dict, contains port and flow.
        for example:
            { "45000", 1234, "45001": 5555 }
        """
        if cmd is None and rmsg[:4] == b'stat':
            stats = rmsg[ 5: ]
            stats = json.loads( stats )
            for key in stats:
                self.accounts[ int( key ) ]['flow'] += stats[ key ]
            self.stater( self.accounts )
            self.stat_count = 0
            return

        # logger.debug( 'recv: {}, cmd: {}'.format(rmsg, cmd) )
        if rmsg != cmd[1]:
            return
        if len(cmd) < 3:
            logger.info( 'Recved Pong from server' )
            self.stat_count = 0
            return

        add = cmd[ 2 ]
        port = cmd[ 3 ]
        logger.info( 'port {} is added: {}'.format(port, add) )
        if add:
            self.accounts[ port ]['status'] = 'added'
        else:
            self.accounts[ port ]['status'] = 'removed'
            # self.accounts[ port ]['flow'] = 0
        # logger.debug( self.accounts )

        



# -*- coding:utf-8 -*-
"""
    sspymgr
    website application for managing shadowsocks service 
"""

__author__ = 'BriFuture'
__version__ = '0.0.13'

# models
from .models import db

from .path_helper import createLogger, createLoggerFile
from .globalfuncs import getRandomCode, convertFlowToByte, formatTime


# -*- coding: utf-8 -*-
import os
basedir = os.path.abspath( os.path.dirname( __file__) )
import yaml
from pathlib import Path
from .path_helper import CONFIG_PATH
import argparse

class Configuration:
    SECRET_KEY = os.environ.get( 'SECRET_KEY' ) or 'ITISAHARDGUESSSTRING'
    SQLALCHEMY_COMMIT_ON_TEARDOWN = True
    
    def __init__(self, *args, **kwargs):
        self.parse()
        pass
    
    def parse( self ):
        if not os.path.exists( CONFIG_PATH ):
            _write_default_config()

        with open( CONFIG_PATH, 'r', encoding='utf-8' ) as f:
            config_obj = yaml.load( f )
            self.__parse_yaml(config_obj)
        ## 命令行参数会覆盖掉配置文件的设置
        self.__parse_sys_argv()
        self.__check_configuration()

    def __parse_yaml(self, yamlObj):
        self.raw_config = yamlObj
        mgrconfig = yamlObj['sspymgr']
        self.email = yamlObj['email']
        self.debug = False
        if 'debugConfig' in mgrconfig:
            if mgrconfig['debugConfig']['enabled']:
                config = mgrconfig['debugConfig']
                self.debug = True
            else:
                config = mgrconfig['config']
        else:
            config = mgrconfig['config']
        self.host = config['host']
        self.port = config['port']
        self.shadowsocks = config.get('shadowsocks', True)
        self.webserver = config.get('webserver', True)

    def __parse_sys_argv(self):
        parser = argparse.ArgumentParser(description='Run SSPYMGR. Support Python3.x only.')

        parser.add_argument("--log_console", 
            help="Logger output will send to stdout.", action="store_true")

        webgroup = parser.add_argument_group('website', 'website options')
        webgroup.add_argument("--webhost", help="Set IP address that web werver listens, \
            for example 0.0.0.0 for IPv4, :: for IPv6", nargs="?")
        webgroup.add_argument("--webport", help="Set port that Website listens, usually 80", nargs="?")

        ssgroup = parser.add_argument_group('shadowsocks', 'integrated shadowsocks opration')
        ssgroup.add_argument("--disable_ss", help="Enable backend shadowsocks server,\
            disable it when developing the program may be helpful", 
            action="store_true")
        ssgroup.add_argument("--only_ss", help="Only start backend shadowsocks server \
            with defined communication method (socket or unix sock file) ", 
            action="store_true")

        args = parser.parse_args()
        if args.webhost is not None:
            self.host = args.webhost
        if args.disable_ss:
            self.shadowsocks = False
        if args.only_ss:
            self.shadowsocks = True
            self.webserver = False

        self.stream_log = args.log_console
        # -w start web gui server
        # -s start shadowsocks server

    def __check_configuration(self):
        ## TODO check host valid

        pass

def _write_default_config( self ):
    """only call this method when needed, or config file may be overrided
    """
    with open( CONFIG_PATH, 'w', encoding='utf-8' ) as f:
        default = """email:
  account: 'your@account.xx'
  password: 'your password'
  host: 'smtp host'
sspymgr:
  debugConfig: 
    enabled: false # true to override following config
    host: '127.0.0.1'
    port: 5050
    shadowsocks: false
  config:
    host: '0.0.0.0'
    port: 80
    shadowsocks: true
"""
        f.write( default )


# -*- coding: utf-8 -*-

import json
from datetime import datetime, timedelta

from .database import db, Account, AccountFlow
from .utils import logger, ssAddr
from .adapter import SSAdapter

class SSController(object):
    """Account: created -> added -> invalid -> removed
    """

    def __init__(self):
        self.stat_counter = 0
        self.stats = None

    def start(self):
        adapter = SSAdapter( self.process_stats )
        try:
            mgr_addr = ssAddr()
            adapter.conn_ser( server = mgr_addr )
        except Exception as exc:
            print( exc )
        self._adapter = adapter
        self.__load_accounts_fromdb()

    def setStats(self, stats):
        self.stats = stats

    def process_stats(self, accounts, force=False):
        if self.stat_counter >= 1 or force:
            if self.stats is not None:
                self.stats(accounts)
            # logger.debug("stats func: {}".format(self.stats))
            self._do_process_stats(accounts)
            self.stat_counter = 0
        self.stat_counter += 1

    def _do_process_stats(self, accounts):
        """@param accounts, send from ss adapter
        """
        dbaccounts = Account.query.all()
        accountFlow = AccountFlow.query.all()
        for acc in dbaccounts:
            p = acc.port
            for flow in accountFlow:
                if p == flow.port:
                    acc._flow = flow
                    break
            logger.debug("process stats: {}".format(accounts[p]))
            if p in accounts.keys() and accounts[p]['flow'] != 0:
                acc._flow.updateTime = datetime.now()
                acc._flow.flow += accounts[p]['flow']
                accounts[p]['flow'] = 0

            if not acc.is_valid():
                acc.totalFlow = 0
                acc._flow.flow = 0
                self._adapter.remove_port(p)
                acc.status = 'removed'
            else:
                if acc.status == 'removed':
                    self._adapter.add_port(p, acc.password)
                    acc.status = 'added'
            # logger.debug("process stats: {}".format(accounts[p]))
            # logger.debug( 'after statis: %d %d' % (acc.port, acc._flow.flow ) )
        db.session.commit()

    def __load_accounts_fromdb(self):
        """only called once after program restart
        """
        dbaccounts = Account.query.all()
        accountFlow = AccountFlow.query.all()
        for acc in dbaccounts:
            for flow in accountFlow:
                if acc.port == flow.port:
                    acc._flow = flow
                    break
            if acc.status == 'removed' or acc.status == 'invalid':
                if acc.is_valid():
                    self._adapter.add_port(acc.port, acc.password)
                    acc.status = 'added'
            elif acc.status == 'added':
                self._adapter.add_port(acc.port, acc.password)

    def add_account(self, port, password, **kwargs):
        """
        @params 
            port: int = None
            password: str = None
            flow in the unit of byte,
            expire 
        @returns bool True if successfully added
        """
        if port in self._adapter.accounts:
            return 'Account already exists'
        # create account
        account = Account(
            port=port,
            password=password,
            userId=kwargs.get('userId'),
            totalFlow=kwargs['flow'],
            expire=kwargs['expire'],
            status='added')
        now = datetime.now()
        flow = AccountFlow(port=port, updateTime=now, checkTime=now)
        flow.autoBanTime = kwargs['expire']

        db.session.add(account)
        db.session.add(flow)
        self._adapter.add_port(port, password)
        db.session.commit()
        return True

    def del_account(self, port, **kwargs):
        """delete an account in ss
        """
        self._adapter.remove_port(port)
        account = Account.query.filter_by(port=port).first()
        account.status = 'removed'
        accountFlow = AccountFlow.query.filter_by(port=port).first()
        db.session.delete(accountFlow)
        db.session.delete(account)
        db.session.commit()

    def get_account(self, id):
        account = Account.query.filter_by(id=id).first()
        if account is not None:
            accountFlow = AccountFlow.query.filter_by(
                port=account.port).first()
            account._flow = accountFlow
        return account

    def renew_account(self, flow, duration, port=None, uid=None):
        """renew an account with new flow and duration
        @param 
            flow: in unit byte
            duration: timedelta
        """
        if port:
            acc = Account.query.filter_by(port=port).first()
        elif uid:
            acc = Account.query.filter_by(userId=uid).first()
        else:
            return False

        # it will cause more time instead of less
        date = datetime.now()
        if date < acc.expire:
            date = acc.expire
        expire = duration + date
        self._update_account(acc, flow, expire)
        return True

    def _update_account(self, acc, flow: int = None, expire: datetime = None):
        """@param acc: account
        """
        accFlow = AccountFlow.query.filter_by(port=acc.port).first()
        # a problem here is that user may lose some flow if last order is active
        accFlow.flow = 0
        if flow is not None:
            acc.totalFlow = flow
        if expire is not None:
            acc.expire = expire
        acc.status = 'added'
        ## communicate with shadowsocks
        self._adapter.remove_port(acc.port)
        self._adapter.add_port(acc.port, acc.password)
        db.session.commit()
        return True

    
    def update_password(self, password, port=0, uid=None):
        """更新 Shadowsocks 账户密码
        若设置了 uid，则优先通过 userId 匹配账户，若同时设置 port 和 uid，则会检查 port 与 uid 是否匹配
        若 uid 和 port 都不设置，则不操作数据库，返回 False
        @returns True 更新成功， False 更新失败 
        """
        if uid is None:
            account = Account.query.filter_by(port=port).first()
        else:
            account = Account.query.filter_by(userId=uid).first()
            if port != 0 and str(account.port) != port:
                return False
        if not account:  # in case account not found
            return False
        self._adapter.update_port(account.port, password)
        account.password = password
        db.session.commit()
        return True

    def execute(self):
        """send command and recieve message
        """
        self._adapter.execute()

    def list_account(self, atype='all'):
        """
        @ param atype
            'all'   list of all accounts
            'valid' list of all valid accounts
            'invalid' list of all invalid accounts
        @returns
            a list with all account
        """
        dbaccounts = Account.query.all()
        accountFlow = AccountFlow.query.all()
        for acc in dbaccounts:
            for flow in accountFlow:
                if acc.port == flow.port:
                    acc._flow = flow
                    break
        return dbaccounts



# -*- coding: utf-8 -*-

__author__ = 'BriFuture'

from gevent import monkey
monkey.patch_all ()

from sspymgr.globalvars import config, events, tasker, \
    logger, controller, emailManager

def init_routes(app):
    # from .routes.vuefront import vuefront
    from .models import db

    @app.teardown_request
    def teardownd_database(resp):
        db.session.close()
        return resp

    from .routes import api
    # 用于添加其他的路由
    app.m_events.trigger('beforeRegisterApi', eventArgs = api )
    from .mail import registerApi as initEmail
    initEmail(api)
    app.register_blueprint( api, url_prefix='/api')


def on_taskSchedule(schedule):
    controller.start()
    schedule.every().seconds.do(controller.execute)
    schedule.every().seconds.do(emailManager.checkRemain)


from flask import Flask

def init_all():
    """initialize all variables and databases.
    return Flask instance app.
    """
    app = Flask( __name__ )
    from .dbsessions import replaceSessionInterface
    replaceSessionInterface(app)

    app.m_events = events
    app.logger = logger
    app.m_emailManager = emailManager
    app.m_sscontroller = controller
    events.on("task_schedule", on_taskSchedule)
    
    # 添加插件机制，要在初始化路由前初始化插件
    from ._plugin import load_plugins
    load_plugins(app)
    
    tasker.start()

    from .models import db
    app.m_events.trigger('beforeCreateDb', eventArgs = db)
    db.create_all()
    init_routes(app)
    
    return app

def run_forever():
    if config.stream_log:
        import sspymgr.path_helper as ph
        ph.log_to_console = True
    if not config.webserver:
        logger.info("Webserver sspymgr webserver not start. Only shadowsocks will running")
        from .sscontroller import start_shadowsocks
        start_shadowsocks(wait=True)
        return

    if config.shadowsocks:
        from .sscontroller import start_shadowsocks
        start_shadowsocks()
    else:
        logger.info("shadowsocks won't start.")
    run_webserver()

def run_webserver():
    app = init_all()

    from gevent import pywsgi
    from .path_helper import createLogger

    weblogger = createLogger("webwsgi")

    if config.debug:
        from werkzeug.debug import DebuggedApplication
        app.debug = True
        app.jinja_env.auto_reload = True
        app.config['TEMPLATES_AUTO_RELOAD'] = True
        dapp = DebuggedApplication( app, evalex= True)
        server = pywsgi.WSGIServer( ( config.host, config.port ), 
            dapp, log = weblogger )
        logger.info( 'starting web server...')
        server.serve_forever()
    else:
        app.debug = False
        server = pywsgi.WSGIServer( ( config.host, config.port ), 
            app, log = weblogger )
        logger.info('starting web server...')
        server.serve_forever()




import hashlib, random, string
code_lib = string.ascii_letters + string.digits

def getRandomCode( bit = 6 ):
    return ''.join( random.SystemRandom().choices( code_lib, k = bit ) )

def convertFlowToByte( flow: int, unit: string ):
    """
    convert flow into bytes
    @param unit: UpperCase, could be G, M, K
    """
    if unit == 'GB' or unit == 'G':
        return flow * 1024 * 1024 * 1024
    elif unit == 'MB' or unit == 'M':
        return flow * 1024 * 1024
    elif unit == 'KB' or unit == 'K':
        return flow * 1024
    
    return flow

def convertByteToFlow( bytes: int ):
    # if bytes < 1024:
    #     return '%d b' % bytes
    # elif bytes < 1024 * 1024:
    #     return '%d Kb' % bytes / 1024
    mb = bytes / 1024 / 1024
    if mb < 1024:
        return '%d Mb' % mb

    gb, mb = divmod( mb, 1024 )
    if mb != 0:
        return '%d Gb %d Mb' % ( gb, mb )
    else:
        return '%d Gb' % gb

from datetime import datetime
def formatTime( dt: datetime ):
    """given datetime object, return string format, 
    second will be hide
    """
    return dt.strftime( '%Y-%m-%d %H:%M' )

def second2str( seconds: int ):
    days, remains = divmod( seconds, 86400 )
    hours, remains = divmod( remains, 3600 )
    minutes, remains = divmod( remains, 60 )
    str = ''
    if days != 0:
        str += "%d days" % days
    if hours != 0:
        str += "%d hours" % hours
    if minutes != 0:
        str += "%d minutes" % minutes
    return str

def mbflow2str( flow: int ):
    """unit of flow is MB
    """
    str = None
    if flow < 1024:
        return '%d Mb' % flow

    gb, mb = divmod( flow, 1024 )
    if mb != 0:
        return '%d Gb %d Mb' % ( gb, mb )
    else:
        return '%d Gb' % gb





# -*- coding: utf-8 -*-

from .models import db
from datetime import datetime

class Email( db.Model ):
    __tablename__ = 'email'
    id = db.Column( db.Integer, primary_key = True, autoincrement = True )
    to = db.Column( db.String( 255 ), nullable=False )
    content = db.Column( db.Text )
    time = db.Column( db.DateTime, default=datetime.now )
    subject = db.Column( db.String( 255 ) )
    ip = db.Column( db.String( 255 ) )
    type = db.Column( db.String( 255 ) )
    remark = db.Column( db.String( 255 ) )
    session = db.Column( db.String( 255 ) )

    def to_dict(self):
        di = {
            'id': self.id,
            'to': self.to,
            'content': self.content,
            'time': self.time.timestamp(),
            'subject': self.subject,
            'ip': self.ip,
            'type': self.type,
            'remark': self.remark,
            'session': self.session,            
        }
        return di

    def __repr__(self):
        return "<Email to: {}, subject: {}, content: {}, time: {}>"\
            .format(self.to, self.subject, self.content, self.time)

from queue import Queue
import smtplib
import yaml
from email import encoders
from email.header import Header
from email.mime.text import MIMEText
from email.utils import parseaddr, formataddr

from .path_helper import createLogger
logger = createLogger('email')

class EmailManager(object):
    def __init__(self, config, *args, **kwargs):
        # self.config = config
        _email = config.email
        self._test = config.debug
        self._host     = _email[ 'host' ]
        self._account  = _email[ 'account' ]
        self._password = _email[ 'password' ]
        self._port     = _email.get( 'port' ) or 465
        if( self._host == 'smtp host' ):
            logger.warning( '[Email] Please Place Your smtp account into config file')
            import sys
            sys.exit( 1 )
        if( self._host == None ) :
            raise ValueError

        self._sendQueue = Queue()

    def add_email(self, to, subject = None, content = None, type = None ):
        email = Email(
            to=to,
            subject=subject,
            content=content,
            type=type,
            remark="pending")
        db.session.add(email)
        self._sendQueue.put(email)
    
    def send(self, email):
        self._sendQueue.put(email)

    def checkRemain(self):

        if self._sendQueue.empty():
            return
        email = self._sendQueue.get()
        db.session.add(email)
        if self._test:
            email.remark = "debugNotSend"
            logger.debug("[Email] Email generated but not send: {}".format(email))
        else:
            ms = MailSender(self._host, self._account, self._password, self._port)
            ms.send(email)
            email.remark = "sent"
        
        db.session.commit()

class MailSender(object):
    def __init__(self, host, account, password, port, *args, **kwargs):
        self.host = host
        self.account = account
        self.password = password
        self.port = port
        self.server = smtplib.SMTP_SSL(self.host, self.port)
        return super().__init__(*args, **kwargs)
    
    def send(self, email: Email):
        msg = MIMEText(email.content, 'html', 'utf-8')
        msg['Subject'] = Header(email.subject, 'utf-8').encode()
        msg['From'] = self.__format_addr('SSPY-Mgr <{}>'.format(self.account))
        msg['To'] = self.__format_addr('Receiver <{}>'.format(email.to) )
        self.server.connect( self.host, self.port )
        # self.server.starttls()
        self.server.ehlo()
        self.server.login( self.account, self.password)
        self.server.sendmail( self.account, email.to, msg.as_string())
        self.server.quit()

    def __format_addr( self, s ):
        name, addr = parseaddr(s)
        return formataddr( ( Header(name, 'utf-8').encode(), addr ) )

def registerApi(api):
    from flask import jsonify 
    @api.route_admin('/email/getAll', methods=['GET', 'POST'])
    def admin_getAllEmails():
        emails = Email.query.all()
        em = [e.to_dict() for e in emails]
        return jsonify({'status': 'success', 'emails': em})

    @api.route_admin('/email/getPage', methods=['GET', 'POST'])
    def admin_getPageEmails():
        page, per_page = api.getPageArgs()
        emails = Email.query.paginate(page=page, per_page=per_page)
        return jsonify({'status': 'success', 'emails': emails})
        


# -*- coding: utf-8 -*-

from .models import db
from flask.sessions import SessionMixin, SessionInterface
from werkzeug import CallbackDict
from uuid import uuid4
from datetime import timedelta, datetime

class Sessions( db.Model ):
    __tablename__ = 'sessions'
    sid = db.Column( db.String( 255 ), primary_key = True )
    expired = db.Column( db.DateTime )
    session = db.Column( db.Text )

    def __init__( self, **kwargs ):
        super( Sessions, self ).__init__( **kwargs )
        if 'sid' in kwargs:
            self.sid = kwargs[ 'sid' ]
        
        if 'expired' in kwargs:
            self.expired = kwargs[ 'expired' ]
        else:
            self.expired = datetime.now() + timedelta( minutes = 10 ) # from now

        if 'session' in kwargs:
            self.session = kwargs[ 'session' ]

    def __repr__( self ):
        return '<Session %s %s %s>' % ( self.sid, self.session, self.expired )

class DbSession( CallbackDict, SessionMixin ):
    def __init__( self, initial = None, sid = None, new = False ):
        def on_update( self ):
            self.modified = True

        CallbackDict.__init__( self, initial, on_update )
        self.sid = sid
        self.new = new
        self.modified = False

# def total_seconds( days, seconds = 0 ):
#     return days * 60 * 60 * 24 + seconds
import json

class DatabaseSessionInterface( SessionInterface ):
    serializer = json
    session_class = DbSession

    def __init__( self, database, prefix = 'session:' ):
        self.db = database
        self.prefix = prefix
        now = datetime.now()
        Sessions.query.filter(now > Sessions.expired).delete()
        db.session.commit()

    def generate_sid( self ):
        return str( uuid4() )
    
    def get_expiration_time( self, app, session ):
        if session.permanent:
            return datetime.now() + app.permanent_session_lifetime
        
        # keep session for 10 minutes
        return datetime.now() + timedelta( minutes = 10 )

    def open_session( self, app, request ):
        sid = request.cookies.get( app.session_cookie_name ) # 尝试从请求中获取存储的 cookie（与 session 相关的 cookie）
        if not sid:
            # 没有从 cookie 中找到相关 sid，说明客户端是第一次访问，需要构造 sid
            sid = self.generate_sid()
            # create new Session representation
            dbrecord = Sessions( sid = sid )
            # db.session.add( dbrecord )
            sc = self.session_class( sid = sid, new = True )
            # print( 'open_new: ', dbrecord, sc.new )
            return sc
        
        # if client have a session_cookie, take relative one from database
        dbrecord = Sessions.query.filter_by( sid = sid ).first()
        
        # print( "open: ", dbrecord )

        if dbrecord is None:
            return self.session_class( sid = sid ) # in case of wrong sid in session_cookie
        else:
            return self.session_class( self.serializer.loads( dbrecord.session ), sid = sid )

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain( app )
        # print( "save ", session, dbrecord )
        # print( 'save new ', session.new )
        dbrecord = Sessions.query.filter_by( sid = session.sid ).first()
        if not session:  
            # session content is empty
            if session.modified:
                # execute deletion if session is empty but modified
                db.session.delete( dbrecord )
                db.session.commit()
                response.delete_cookie( app.session_cookie_name, domain = domain )

            return 
        
        # update or insert
        if dbrecord is None:
            # in case database is clean
            dbrecord = Sessions( sid = session.sid )
            db.session.add( dbrecord )
            
        dbrecord.session = self.serializer.dumps( session )
        sexp = self.get_expiration_time( app, session )
        dbrecord.expired = sexp
        # print( "save:", dbrecord )
        db.session.commit()
        response.set_cookie( app.session_cookie_name, session.sid,
            expires = sexp, httponly = True, domain = domain )

import hashlib
from .globalfuncs import getRandomCode
def replaceSessionInterface(app):
    app.session_interface = DatabaseSessionInterface( database = db )
    # app.secret_key = hashlib.sha1( getRandomCode( 24 ).encode() ).hexdigest()
    app.secret_key = hashlib.sha1( 'test'.encode() ).hexdigest()


import sys, os

from sspymgr.path_helper import createLoggerFile, DATA_DIR
from sspymgr.sscontroller import ssAddr

def ssConfigFile():
    config = DATA_DIR + '/server-multi-passwd.json'
    if( not os.path.exists( config ) ):
        
        DEFAULT_CONFIG = """{
    "server": "0.0.0.0",
    "server_port": 0,
    "timeout": 300,
    "method": "aes-256-cfb",
    "fast_open": false
}
"""
        with open( config, 'w', encoding="utf-8") as f:
            f.write( DEFAULT_CONFIG )
    return config


def runSS():
    mgr_addr = ssAddr(True)
    args = '--manager-address %s -c %s' % ( mgr_addr, ssConfigFile() )
    print(sys.argv)
    sys.argv[1:] = args.split( ' ' )   
    # import logging
    # logging.basicConfig(level=logging.INFO,
    #     format='%(asctime)s %(levelname)-8s %(message)s',
    #     datefmt='%Y-%m-%d %H:%M:%S')
    from shadowsocks import server
    server.main()

def new_runSS():
    import logging
    fileHandler = createLoggerFile('shadowsocks')
    logger = logging.getLogger()
    logger.addHandler(fileHandler)
    logger.setLevel(logging.INFO)

    from shadowsocks import shell, daemon, eventloop, manager
    logger.info( 'Start running shadowsock server')
    shell.check_python()
    logging.basicConfig(level=logging.INFO,
        format='%(asctime)s %(levelname)-8s %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')

    config_path = ssConfigFile()
    logger.info('loading config from %s' % config_path)
    with open(config_path, 'rb') as f:
        try:
            config = shell.parse_json_in_str(f.read().decode('utf8'))
        except ValueError as e:
            logger.error('found an error in config.json: %s',
                            e.message)
            sys.exit(1)
    config['manager_address'] = ssAddr(True)
    config['port_password'] = {}
    config['server_port'] = 0
    config['verbose'] = 2
    config['method'] = 'aes-256-cfb'

    daemon.daemon_exec(config)

    config['port_password'] = {}
    # server_port = config['server_port']
    # if type(server_port) == list:
    #     for a_server_port in server_port:
    #         config['port_password'][a_server_port] = config['password']
    # else:
    #     config['port_password'][str(server_port)] = config['password']

    logger.info('entering manager mode')
    manager.run(config)

def main(wait=False):
    # import logging
    # log = logging.getLogger( 'Test' )
    # import sys
    # from subprocess import Popen
    # with open( 'logfile.txt', 'ab' ) as file:
        # p = Popen( ['python', '-m', 'sockserver' ], stdin=sys.stdin, stdout=sys.stdout, stderr=file  ) 
    
    from multiprocessing import Process
    p = Process( target = new_runSS )
    p.daemon = True
    p.start()
    if wait:
        p.join()
    from time import sleep
    sleep( 1 )

if __name__ == '__main__':
    startSS()



# -*- coding: utf-8 -*-

from .path_helper import createLogger
logger = createLogger('pluginLoader', stream=False, logger_prefix="[PluginLoader]")

class Events(dict):
    def __getitem__(self,name):
        if name not in self:
            super().__setitem__(name, [])
        return super().__getitem__(name)

    def trigger(self, eventName: str, eventArgs = None, **kwargs):
        """trigger event `eventName`
        """
        try:
            for e in self[eventName]:
                e(eventArgs, **kwargs)
        except Exception as exc:
            logger.warn("Exception at 'beforeRegisterRoutes: {}".format(exc))
        
    def on(self, eventName: str, callback ):
        """append callback function for certain event
        """
        self[eventName].append(callback)

import schedule

class BackTask(object):
    def __init__(self, events):
        self.events = events

    def start(self):
        """this method makes sure that endless while loop running in a subthread and will not interfere webgui
        """
        from threading import Thread
        
        mt = Thread( target= self.__loop_monitor )
        mt.daemon = True
        mt.start()
        
    def __loop_monitor(self):
        import time
        self.events.trigger("task_schedule", schedule)
        logger.info( 'start background tasks....' )

        while True:
            schedule.run_pending()
            time.sleep( 1 )


def load_plugins(app):
    from pathlib import Path
    import os
    pluginPath = Path(os.path.dirname(__file__)) / "plugins"
    
    import sys
    sys.path.insert(0, str(pluginPath))

    # 过滤掉非插件文件
    plugins = []
    for file in pluginPath.iterdir():
        if file.name.startswith('__'):
            continue
        if file.is_dir():
            dir_content = [x.name for x in file.iterdir()]
            if "__init__.py" in dir_content:
                plugins.append(file)
            dir_content.clear()
        elif file.is_file() and file.suffix == ".py":
            plugins.append(file)
    
    import importlib

    # 过滤掉依赖关系无法满足的插件
    logger.info("开始过滤插件。")
    plugin_module = []
    no_dependences = []
    for plugin_file in plugins:
        try:
            name = plugin_file.stem
            # print(' ==>', name)
            if plugin_file.is_dir():
                # following works for python3.6
                pm = importlib.import_module("{}".format(name))
                # following works for python3.7
                # pm = importlib.import_module("{}/__init__".format(name))
            else:
                pm = importlib.import_module(name)
            pm.name = name
            logger.debug("导入插件 {}".format(pm))

            if(not hasattr(pm, 'requirement') or len(pm.requirement) == 0):
                no_dependences.append(pm)
            else:
                meet = True
                for r in pm.requirement:
                    if r not in plugins:
                        meet = False
                if meet:
                    plugin_module.append(pm)
        except Exception as exc:
            logger.warn("Load plugin `{}` Error: {}".format(plugin_file, exc))
    plugins.clear()
    importlib.invalidate_caches()
    logger.info("过滤插件完毕，无依赖插件数：{}，有依赖插件数: {}"\
        .format(len(no_dependences), len(plugin_module)))

    num_all_plugins = len(no_dependences) + len(plugin_module)
    logger.info("准备初始化插件")
    app.m_plugins = []
    num_loaded = 0
    level_loaded = 0
    test = 0
    while(num_loaded < num_all_plugins and level_loaded < 3):
        num_loaded += __do_load_plugin(app, plugin_module, no_dependences)
        level_loaded += 1
    logger.info("插件初始化完成, 加载了{}个插件".format(num_loaded))

def __do_load_plugin(app, plugin_module, no_dependences):
    # 首先加载没有依赖的插件
    # logger.debug("no_dependces: {}".format(no_dependences))
    num_loaded = 0
    for nd in no_dependences:
        try:
            logger.debug("初始化插件: {}".format(nd))
            nd.init(app)
            # logger.debug("load: {}".format(nd))
            app.m_plugins.append(nd)
            num_loaded += 1
        except Exception as exc:
            logger.warn("Init plugin `{}` Error: {}".format(nd, exc))

    # logger.debug("""loaded: {}, \nno_depend: {}, 
    #     plugin_module: {}""".format(app.m_plugins, no_dependences, plugin_module))
    no_dependences.clear()

    # 满足依赖关系的插件
    for depend in plugin_module:
        for pm in app.m_plugins:
        # for req in pm.requirement:
            if pm.name in depend.requirement:
                depend.requirement.remove(pm.name)

    no_dependences.clear()
    tmp_plugins = []
    for pm in plugin_module:
        if len(pm.requirement) > 0:
            tmp_plugins.append(pm)
        else:
            no_dependences.append(pm)
    plugin_module.clear()
    for pm in tmp_plugins:
        plugin_module.append(pm)
    # logger.debug("no_depend: {}".format(no_dependences))
    return num_loaded
            



# -*- coding: utf-8 -*-


from sspymgr import createLogger
logger = createLogger(
    "plugin_sscontroller", stream=False, logger_prefix="[Plugin sscontroller]")
from .ssc_routes import registerApi, availiable_port
from sspymgr.globalvars import config, controller
from sspymgr import getRandomCode
from datetime import datetime, timedelta

from .flow_stats import FlowStats, checkAllAccounts
def scheduleTasks(schedule):
    recorder = FlowStats()
    controller.setStats(recorder.recordFlow)
    for i in range(1, 13):
        schedule.every(1).hours.at(":{:02d}".format(i*5-1)).do(recorder.record5minToDb)
    schedule.every().day.at("23:59").do(recorder.record1dayToDb)
    schedule.every().minutes.do(checkAllAccounts)
    logger.debug("Tasks scheduled done")

def afterSingup(params: dict):
    port = availiable_port()
    password = getRandomCode(8)
    from web_settings import WebguiSetting
    days = WebguiSetting.getSetting(key='web_signup_trial_days',\
            default_value=1, default_type="Number").getTypedValue()
    expire = timedelta(days=days) + datetime.now()
    added = controller.add_account(
        port,
        password,
        userId=params['userId'],
        flow=params['flow'],
        expire=expire
    )
    logger.debug("signup accounts added: {}, params: {}".format(added, params))

def init(app):
    app.m_events.on('beforeRegisterApi', registerApi)
    app.m_events.on('task_schedule', scheduleTasks)
    app.m_events.on('webuser_signup_success', afterSingup)
    logger.info("inited")


# -*- coding: utf-8 -*-

from sspymgr import db, createLogger
from sspymgr import getRandomCode, convertFlowToByte

from flask import session
from datetime import datetime, timedelta
import hashlib

logger = createLogger("plugin_user", stream=False, logger_prefix="[Plugin User]")

class User(db.Model):
    __tablename__ = 'user'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    type = db.Column(db.String(255))
    salt = db.Column(db.String(8), nullable=False)
    email = db.Column(db.String(255), nullable=False)
    group = db.Column(db.Integer, default=0)
    level = db.Column(db.Integer, default=1)
    comment = db.Column(db.String(255), default='')
    password = db.Column(db.String(255), nullable=False)
    telegram = db.Column(db.String(255))
    username = db.Column(db.String(255))
    lastLogin = db.Column(db.DateTime)
    createTime = db.Column(db.DateTime)
    resetPasswordId = db.Column(db.String(255))
    resetPasswordTime = db.Column(db.DateTime)

    def __init__(self, **kwargs):
        super(User, self).__init__(**kwargs)
        # self.password = password
        self.salt = getRandomCode(8)
        self.password = self.__gen_passwd(self.password)
        self.createTime = datetime.now()

    def __repr__(self):
        return '<User %s %s %s>' % (self.email, self.salt, self.password)

    def __gen_passwd(self, password):
        """
        Just generate secure password, but not modify object's password attribute,
        parameter password can't be None or self.password, otherwise the method 
        which checks whether password is valid will return a Fake result
        """
        return hashlib.sha256((self.salt + password).encode()).hexdigest()

    def updatePassword(self, new_passwd):
        self.password = self.__gen_passwd(new_passwd)

    def isPasswordValid(self, password):
        """判断用户密码是否正确 
        """
        password = str(password)
        password = self.__gen_passwd(password)
        return password == self.password

    def request_reset_passwd(self):
        self.resetPasswordId = getRandomCode(24)
        self.resetPasswordTime = datetime.now()

    def do_reset_passwd(self, passwdId, new_passwd):
        """reset password and clear requestId
        @returns
            True if succesfully reset, False otherwise
        """
        if not self.resetPasswordId or not self.resetPasswordTime:
            return False
        now = datetime.now()
        diff = now - self.resetPasswordTime

        if diff > timedelta(hours=3):
            # overtime, invalid
            return False

        if passwdId != self.resetPasswordId:
            return False

        self.password = self.__gen_passwd(new_passwd)
        self.resetPasswordId = ''
        return True

    def to_dict(self):
        lastLogin = self.lastLogin.timestamp() if self.lastLogin is not None else None
        di = {
            'id': self.id,
            'email': self.email,
            'type': self.type,
            'group': self.group,
            'comment': self.comment,
            'telegram': self.telegram,
            'username': self.username,
            'lastLogin': lastLogin,
            'createTime': self.createTime.timestamp(),
        }
        return di


class Group(db.Model):
    __tablename__ = 'group'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    name = db.Column(db.DateTime)
    order = db.Column(db.String(255))
    comment = db.Column(db.DateTime)
    showNotice = db.Column(db.Integer, default=1)
    multiAccount = db.Column(db.Integer, default=0)


from enum import Enum, unique


@unique
class UserType(Enum):
    BANNED = 'banned'
    TRIAL = 'trial'
    ACTIVE = 'active'
    INVALID = 'invalid'
    NOTIFIED = 'notified'
    DEFAULT = 'default'
    TEST = 'test'


# class UserStatus(db.Model):
#     __tablename__ = 'UserStatus'
#     userId = db.Column(db.Integer, nullable=False)
#     webStatus = db.Column(db.String(64), default=UserType.DEFAULT.value)
#     ssStatus = db.Column(db.String(64))


def getCurrentUser():
    uid = session.get('userid')
    user = User.query.filter_by(id=uid).first()
    return user


def isManager(user=None):
    if user is not None:
        return user.id == 1
    if 'userid' in session:
        return session.get('userid', -1) == 1
    return False


def getSuperManager():
    return User.query.filter_by(id=1).first()


def user_signin(user):
    """process user signin action, assume user is validated
    """
    session['user'] = user.email
    session['userid'] = user.id
    user.lastLogin = datetime.now()


def send_checkcode_email(emailaddr, emailManager):
    """注册时获取验证码
    """
    rcode = getRandomCode()
    session['usercheckcode'] = rcode
    content = '你好，注册验证码为: %s' % rcode
    emailManager.add_email(
        to=emailaddr,
        subject='SSPY SignUp',
        content=content,
        type="signup checkcode")

def send_reset_link(host, email, code, emailManager):
    """重置密码
    """
    link = host + '/vhome#/resetPassword/' + code
    # url_for( 'home.on_resetPasswdWithId', code = code)
    content = 'Hi, 重置密码的链接为: %s，请在时限内重置你的密码' % link
    emailManager.add_email(to=email, subject="SSPY Reset Password Link",
        content=content, type="reset password")


from flask import request, jsonify

def init(app):
    app.m_events.on('beforeRegisterApi', registerApi)
    logger.debug("inited")

MAX_CHECKCODE_ATTEMPT = 3
CHECKCODE_TIMEOUT = timedelta(minutes=10)

def registerApi(api):
    from sspymgr.globalvars import EMAIL_REGEX, emailManager, events
    from web_settings import WebguiSetting

    @api.route('/home/signin', methods=['POST'])
    def on_home_signin():
        """Basic entry to signin for users who want to use sspymgr
        """
        # first of all, validate input data
        resp = {'status': 'success'}
        emailaddr = request.form.get('email', '')
        password = request.form.get('password', '')
        valid = True
        if (len(emailaddr) == 0):
            resp['email'] = 'Email is Empty'
            valid = False
        if (len(password) == 0):
            resp['password'] = 'Password is Empty'
            valid = False

        if valid is False:
            resp['status'] = 'fail'
            return jsonify(resp)
        # find correspoding user and if not found, raise error
        user = User.query.filter_by(email=emailaddr).first()
        if user is None:
            resp['status'] = 'fail'
            resp['email'] = 'Email is not valid'
            return jsonify(resp)

        if user.isPasswordValid(password):
            # able to sign in
            user_signin(user)
            session.permanent = request.form.get(
                'keep_signedin') == 'remember-me'
            resp['status'] = 'success'
            resp['level'] = 'admin' if isManager(user) else 'user'
        else:
            # unable to signin, password may be wrong
            resp['status'] = 'fail'
            resp['password'] = 'Password is wrong'
        return jsonify(resp)

    @api.route('/home/isSignedIn', methods=['POST'])
    def is_user_signedIn():
        resp = {'signedIn': False, 'level': 'user'}
        if 'userid' in session:
            resp['signedIn'] = True
            if isManager():
                resp['level'] = 'admin'

        return jsonify(resp)


    @api.route('/home/checkcode', methods=['POST'])
    def on_send_checkcode():
        resp = {'status': 'success'}
        email = request.form.get('email', '')
        if len(email) == 0:
            resp['status'] = 'fail'
            resp['desc'] = 'Email is empty'
            return jsonify(resp)

        exists = User.query.filter_by(email=email).count() > 0
        if not EMAIL_REGEX.match(email) or exists:
            resp['status'] = 'fail'
            resp['desc'] = 'Wrong Email or Email already in use'
            return jsonify(resp)

        if 'checkcode_attempt' not in session:
            session['checkcode_attempt'] = 0
            session['checkcode_start'] = datetime.now().timestamp()

        singup_limit = WebguiSetting.getSetting('signup_remain_limit',\
            default_value=4, default_type="Number").getTypedValue()
        if singup_limit == 0:
            resp['status'] = 'fail'
            resp['desc'] = 'Sorry, no more users can signup.'
            return jsonify(resp)

        session['checkcode_attempt'] += 1
        
        if session['checkcode_attempt'] <= MAX_CHECKCODE_ATTEMPT:
            send_checkcode_email(email, emailManager)
        else:
            start = datetime.fromtimestamp(session['checkcode_start'])
            now = datetime.now()
            if now - start > CHECKCODE_TIMEOUT:
                session['checkcode_attempt'] = 1
                send_checkcode_email(email, emailManager)
            else:
                resp['status'] = 'fail'
                resp['desc'] = 'Please wait a few minutes before resend check code.'
        db.session.commit()
        return jsonify(resp)


    @api.route('/home/signup', methods=['POST'])
    def on_home_signup():
        resp = {'status': 'fail'}
        valid = True
        if (not 'usercheckcode' in session):
            resp['code'] = 'Get checkcode first.'
            valid = False

        email = request.form.get('email', '')
        checkcode = request.form.get('checkcode', '')
        password = request.form.get('password', '')
        if (len(email) == 0):
            resp['email'] = 'Email is Empty'
            valid = False
        if (len(checkcode) == 0):
            resp['code'] = 'CheckCode is Empty'
            valid = False
        if (len(password) == 0):
            resp['password'] = 'Password is Empty'
            valid = False
        if valid and not EMAIL_REGEX.match(request.form['email']):
            resp['email'] = 'Wrong email or not able to sign up now'
            valid = False

        if valid is False:
            return jsonify(resp)

        user = User.query.filter_by(email=email).first()
        if user is not None:
            resp['status'] = 'fail'
            resp['email'] = 'Email has already in use'
            return jsonify(resp)

        if (session.get('usercheckcode').lower() == checkcode.lower()):
            email = email.strip()
            user = User(email=email, password=password)
            db.session.add(user)
            setting = WebguiSetting.getSetting(key='signup_remain_limit')
            setting.value = setting.getTypedValue() - 1
            db.session.commit()
            # add account into shadowsocks server, but there is a problem is that if not committed,
            # user.id will be None
            user = User.query.filter_by(email=email).first()
            events.trigger("webuser_signup_success", dict(
                userId=user.id,
                flow=convertFlowToByte(1, 'G'),
            ))

            # checkcode will be removed after user have been added
            session.pop('usercheckcode')
            # auto signin
            resp['status'] = 'success'
            user_signin(user)
            
            db.session.commit()
        return jsonify(resp)

    @api.route('/home/requestReset', methods=['POST'])
    def on_home_requestReset():
        state = {'status': 'success'}
        email = request.form.get('email', '')
        if (len(email) == 0 or not EMAIL_REGEX.match(request.form['email'])):
            state['status'] = 'fail'
        user = User.query.filter_by(email=email).first()

        if user:
            user.request_reset_passwd()
            code = '%d_%s' % (user.id, user.resetPasswordId)
            send_reset_link(request.host, email, code, emailManager)
            db.session.commit()
        else:
            state['status'] = 'fail'
        return jsonify(state)

    @api.route('/home/resetPassword', methods=['POST'])
    def on_home_doResetPasswdWithId():
        code = request.form.get('code', '')
        code = code.split('_')
        state = {'status': 'success'}
        if len(code) < 2:
            state['status'] = 'fail'
            state['hint'] = 'Wrong code'
            return jsonify(state)
        userId = code[0]
        resetId = code[1]
        user = User.query.filter_by(id=userId, resetPasswordId=resetId).first()
        newpw = request.form.get('newPassword', '')
        if user is None or len(newpw) == 0:
            state['status'] = 'fail'
            state['hint'] = 'Wrong Password'
            return jsonify(state)

        result = user.do_reset_passwd(resetId, newpw)
        db.session.commit()
        if not result:
            state['status'] = 'fail'
            state['hint'] = 'Your code seems not right or expired'
        return jsonify(state)

    # =====================  user route =====================
    @api.route_user('/signout', methods=['POST'])
    def user_signout():
        session.pop('user')
        session.pop('userid')
        return jsonify('success')

    @api.route_user('/updatePassword', methods=['POST'])
    def on_user_update_password():
        uid = session.get('userid')
        user = User.query.filter_by(id=uid).first()
        oldp = request.form.get('oldPassword')
        if user is None or not user.isPasswordValid(oldp):
            return jsonify('fail')
        newp = request.form.get('newPassword')
        user.updatePassword(newp)
        db.session.commit()
        return jsonify('success')

    # ===============  admin route  =================
    @api.route_admin('/user/updateInfo', methods=['POST'])
    def admin_updateUserInfo():
        comment = request.form.get('comment')
        utype = request.form.get('type')
        uid = request.form.get('userId')
        user = User.query.filter_by(id=uid).first()
        if user is None:
            return jsonify('fail')
        user.comment = comment
        user.type = utype
        db.session.commit()
        return jsonify('success')

    @api.route_admin('/user/getAll', methods=['POST'])
    def get_admin_user():
        users = User.query.filter().all()
        return jsonify({
            'status': 'success',
            'users': [u.to_dict() for u in users]
        })

    logger.debug("api registered.")